import numpy as np


def eucl_dist(u, v, alpha=2.0):
    return ((u - v) ** alpha).sum(axis=1)


def calc_dist(X, Z, func_dist=eucl_dist, **kwargs):

    # prepare the left and right for distances
    u = np.repeat(X, Z.shape[0], axis=0)
    v = np.tile(Z, (X.shape[0], 1))

    # calculate the distances and reshape to distance metric
    D = func_dist(u, v, **kwargs)
    D = np.reshape(D, (X.shape[0], Z.shape[0]))

    return D




import numpy as np

try:
    from sklearn.ensemble import RandomForestRegressor
except:
    raise Exception("Model not found. Please execute: 'pip install sklearn'")

from ezmodel.core.model import Model
from ezmodel.util.misc import discretize


class RandomForest(Model):

    def __init__(self, n_partitions=5, n_estimators=30, xl=None, xu=None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.xl, self.xu = xl, xu
        self.n_partitions = n_partitions
        self.n_estimators = n_estimators

    def _fit(self, X, y, **kwargs):

        if self.xl is None:
            self.xl = X.min(axis=0)
        if self.xu is None:
            self.xu = X.max(axis=0)

        y = y[:, 0]

        X = discretize(X, self.n_partitions, self.xl, self.xu)

        D = {}
        for i, x in enumerate(X):
            s = str(x)
            if s not in D:
                D[s] = dict(X=x, y=y[i], n=1)
            else:
                _y, _n = D[s]['y'], D[s]['n']
                D[s] = dict(X=x, y=min(_y, y[i]), n=_n + 1)

        X = np.zeros((len(D), X.shape[1]))
        y = np.zeros(len(D))
        for i, e in enumerate(D.values()):
            X[i] = e["X"]
            y[i] = e["y"]

        rf = RandomForestRegressor(n_estimators=self.n_estimators, random_state=42)
        rf.fit(X, y)

        self.model = rf

    def _predict(self, X, out):
        out["y"] = self.model.predict(discretize(X, self.n_partitions, self.xl, self.xu))[:, None]

    @classmethod
    def hyperparameters(cls):
        return {
            "n_partitions": [20, 50, 100],
            "n_estimators": [10, 100, 200]
        }



from ezmodel.util.aggregate.clearing import MinSelection
from ezmodel.util.misc import discretize


def aggregate_by_grid(X,
                      n_partitions,
                      selection=MinSelection,
                      return_cluster=False,
                      problem=None,
                      xl=None,
                      xu=None,
                      **kwargs):

    if problem is not None:
        xl, xu = problem.bounds()

    _X = discretize(X, n_partitions, xl=xl, xu=xu)

    D = {}
    for i, x in enumerate(_X):
        s = str(x)
        if s not in D:
            D[s] = [i]
        else:
            D[s] = D[s] + [i]

    ret = {}
    for group in D.values():
        i = selection(problem=problem, **kwargs).do(group)
        ret[i] = [e for e in group if e != i]

    I = list(ret.keys())

    if return_cluster:
        return I, ret
    else:
        return I



import numpy as np


from ezmodel.core.model import Model
from ezmodel.util.dist import calc_dist


class KNN(Model):

    def __init__(self, n_nearest=5, p=1.0, **kwargs) -> None:
        super().__init__(**kwargs)
        self.n_nearest = n_nearest
        self.p = p

    def _fit(self, X, y, **kwargs):
        pass

    def _predict(self, X, out):
        _X, _y = self.X, self.y[:, 0]

        D = calc_dist(X, _X)

        I = D.argsort(axis=1)[:, :self.n_nearest]

        _d = np.take_along_axis(D, I, axis=1)
        _d = (_d ** self.p)

        _d[_d == 0] = 1e-64

        _d = 1 / _d
        _d = _d / _d.sum(axis=1)[:, None]

        _y = np.take_along_axis(self.y, I, axis=0)

        y = (_d * _y).sum(axis=1)

        out["y"] = y[:, None]

    @classmethod
    def hyperparameters(cls):
        return {
            "n_nearest": [3, 5, 10],
            "p": [0.0, 1.0, 2.0, 5.0]
        }




try:
    from pySOT.surrogate import RBFInterpolant, CubicKernel, LinearTail, TPSKernel, ConstantTail
except:
    raise Exception("Model not found. Please execute: 'pip install pySOT'")





from ezmodel.core.model import Model


class pySOTRBF(Model):

    def __init__(self, kernel="cubic", tail="linear", eta=1e-6, **kwargs) -> None:
        super().__init__(**kwargs)
        self.kernel = kernel
        self.tail = tail
        self.eta = eta

    def _fit(self, X, y, **kwargs):
        n, m = X.shape
        kernel, tail = get_kernel(self.kernel), get_tail(self.tail, m)
        self.model = RBFInterpolant(dim=m, kernel=kernel, tail=tail, eta=self.eta)
        self.model.add_points(X, y)

    def _predict(self, X, out):
        out["y"] = self.model.predict(X)

    @classmethod
    def hyperparameters(cls):
        return {
            "kernel": ["cubic", "tps"],
            "tail": ["linear"],
            "eta": [1e-12, 1e-6, 1e-4, 1e-2],
        }


def get_kernel(kernel):
    if kernel == "cubic":
        return CubicKernel()
    elif kernel == "tps":
        return TPSKernel()


def get_tail(tail, m):
    if tail == "linear":
        return LinearTail(m)
    elif tail == "constant":
        return ConstantTail(m)



import numpy as np
from scipy.spatial.distance import cdist


def all_except(x, *args):
    if len(args) == 0:
        return x
    else:
        H = set(args) if len(args) > 5 else args
        I = [k for k in range(len(x)) if k not in H]
        return x[I]


def from_dict(D, *keys):
    return [D.get(k) for k in keys]


def is_duplicate(X, eps=1e-16):
    D = cdist(X, X)
    D[np.triu_indices(len(X))] = np.inf

    I = np.full(len(X), False)
    I[np.any(D < eps, axis=1)] = True
    return I


def to_list(v):
    if not isinstance(v, list):
        v = [v]
    return v


def at_least2d(x, expand="c"):
    if x.ndim == 1:
        if expand == "c":
            return x[:, None]
        elif expand == "r":
            return x[None, :]
    else:
        return x


def empty_dict_if_none(x):
    if x is None:
        return {}
    else:
        return x


def dict_to_str(vals, delim=",", sep="="):
    return delim.join(f'{key}{sep}{value}' for key, value in vals.items())


def discretize(X, n_partitions, xl=None, xu=None):
    if xl is None:
        xl = X.min(axis=0)

    if xu is None:
        xu = X.max(axis=0)

    thresholds = np.linspace(xl, xu, n_partitions + 1)[1:]
    ret = (X[..., None] < thresholds.T).argmax(axis=-1)
    return ret



import copy
import warnings
from multiprocessing.pool import ThreadPool

import numpy as np

from ezmodel.util.metrics import calc_metric
from ezmodel.util.misc import at_least2d
from ezmodel.util.partitioning.crossvalidation import CrossvalidationPartitioning


class Benchmark:

    def __init__(self,
                 models,
                 metrics=["mse", "mae", "r2", "spear"],
                 show_warnings=False,
                 raise_exception=False,
                 n_threads=None,
                 verbose=False):

        """

        This is a benchmark class which evaluates the performance of surrogates on a data set.

        Parameters
        ----------
        models : list
            A list of dictionaries where the keys `label` and `obj` need to be set.

        metrics : list
            What metrics shall be calculated during the benchmark.

        show_warnings : bool
            Whether warnings should be displayed or not. This can cause to show all warnings from
            all surrogates.

        raise_exception : bool
            Whether an exception should be raised if a surrogate fails.

        n_threads : int
            If multi-threading should be used this can be enabled, then each process of fitting is
            parallelized.

        verbose : bool
            Whether the benchmark should provide printouts or not.

        """

        super().__init__()
        self.models = dict(models)

        # the models can either be provided in a dict or directly -> then this reformatting is necessary
        for k, v in self.models.items():
            if not isinstance(v, dict):
                self.models[k] = dict(model=v)
            self.models[k]["label"] = k

        # the metrics to be used to evaluate the performance of each model
        self.metrics = metrics

        # whether warnings shall be displayed
        self.show_warnings = show_warnings

        # whether an exception shall be raised if a model fails
        self.raise_exception = raise_exception

        # the number of threads, None simply disables multi-threading
        self.n_threads = n_threads

        # if true, the benchmark does some printouts
        self.verbose = verbose

        # the records when the benchmark has been executed
        self.records = None

        # the overall results grouped by each surrogate
        self.records_by_model = None

        # storage for data for each experiment
        self.data = None

    def do(self, X, y, partitions=None):
        X, y = at_least2d(X, expand="r"), at_least2d(y, expand="c")[:, 0]
        assert len(X) == len(y)

        # when no partitions are provided by default a k-fold cross-validation is done
        if partitions is None:
            partitions = CrossvalidationPartitioning(X, 3, seed=1).do()

        # store the current setup of the benchmark
        self.data = dict(X=X, y=y, partitions=partitions)

        # an empty list where the records are added to
        self.records = []

        # create all the entries each of them representing a job for fitting a model - for each partition
        for k in range(len(partitions)):

            # for each model to be tested
            for _, entry in self.models.items():
                model = entry["model"]
                val = dict(
                    benchmark="benchmark@%s" % id(self),
                    proto=model,
                    partition=k,
                )
                self.records.append({**entry, **val})

        # fit all the models for each partition
        self._do()

        # evaluate the performance for each of the methods
        self._performance()

        # do some more post-processing to prepare the results
        self._postprocessing()

        return self

    def _do(self):
        if self.n_threads is None:
            ret = [fit(self.data, entry, self.show_warnings, self.raise_exception) for entry in self.records]
        else:
            with ThreadPool(self.n_threads) as pool:
                ret = pool.starmap(fit,
                                   [(self.data, entry, self.show_warnings, self.raise_exception) for entry in
                                    self.records])
        return ret

    def _performance(self):

        # the true data for predictions
        partitions, y_true = self.data["partitions"], self.data["y"]

        # now set the results based on the predictions
        for record in self.records:

            # set all metrics to none which ensures they are set if an error occurs
            record["trn_error"] = np.nan
            for metric in self.metrics:
                record[metric] = np.nan

            # if the model was fitted successfully
            if record["success"]:

                try:

                    trn, tst = partitions[record["partition"]]
                    y_hat, trn_y_hat = record["y_hat"], record["trn_y_hat"]

                    record["trn_error"] = calc_metric("mae", y_true[trn], trn_y_hat)

                    for metric in self.metrics:
                        warnings.simplefilter("ignore")
                        record[metric] = calc_metric(metric, y_true[tst], y_hat)

                    if self.verbose:
                        print(record["label"], record["obj"].time)

                except:
                    pass

    def _postprocessing(self):
        records = self.records

        # now let us create the results of the benchmark
        results = dict(self.models)
        for key in results.keys():
            results[key] = dict(runs=[], performance={}, **results[key])

        for entry in records:
            hash = entry["label"]
            results[hash]["runs"].append(entry)

        collect = lambda k, x: np.array([run[x] for run in results[k]["runs"]])

        # collect the data from all the runs
        for metric in self.metrics:
            for k, v in results.items():
                vals = collect(k, metric)
                v["performance"][metric] = dict(mean=vals.mean(), std=vals.std(), min=vals.min(), max=vals.max(),
                                                values=vals)

        for k, v in results.items():
            v["n_runs"] = len(v["runs"])
            v["success"] = np.all(collect(k, "success"))
            v["trn_error"] = np.max(collect(k, "trn_error"))

        self.records_by_model = results

    def results(self,
                only_successful=True,
                as_list=True,
                sorted_by=None,
                max_trn_error=None,
                include_metadata=False,
                include_records=False):

        ret = {}
        for k, v in self.records_by_model.items():
            if not only_successful or v["success"]:
                ret[k] = v
                if sorted_by is not None:
                    metric, value, _ = sorted_by
                    v["metric"] = v["performance"][metric][value]

        # if not a list should be returned that's it
        if as_list:

            # otherwise get a list and do some additionally work
            ret = list(ret.values())

            if sorted_by is not None:
                metric, value, ascending = sorted_by

                # sort the indices by their performance
                sign = 1 if ascending else -1
                ret = sorted(ret, key=lambda x: sign * x["metric"])
            else:
                ret = sorted(ret, key=lambda x: x["performance"]["mae"]["mean"])

            # filter out solutions with too much error on training set (for approximation models)
            if max_trn_error is not None:

                is_valid = [e["trn_error"] < max_trn_error for e in ret]

                if not any(is_valid):
                    print(f"WARNING: No model with maximum training error of {max_trn_error} was found.")

                else:
                    ret = [ret[i] for i in range(len(ret)) if is_valid[i]]

        if include_metadata:
            ret = {**dict(self.data), "results": ret}

            if include_records:
                ret["records"] = self.records

        return ret

    def statistics(self, metric="mae", vals=['mean', 'std', 'min', 'max', 'median'], sort_by="mean", ascending=True):
        try:
            import pandas as pd
        except:
            raise Exception("Please install the pandas toolbox for using statistics: pip install pandas")

        df = pd.DataFrame(self.records)
        tbl = df.groupby(["label"]).agg({metric: vals})
        df.groupby(["label"]).agg({metric: vals})

        if sort_by is not None:
            tbl = tbl.sort_values((metric, sort_by), ascending=ascending)

        return tbl

    def correlation(self, models=None):

        if models is None:
            models = [model for model, vals in self.data.items() if vals["success"]]

        preds = {}
        for model in models:
            preds[model] = np.concatenate([run["y_hat"] for run in self.data[model]["runs"]])

        import pandas as pd
        return pd.DataFrame(preds).corr()


# ----------------------------------------------------------------
# Util
# ----------------------------------------------------------------

def fit(data, record, show_warning, raise_exception):
    try:

        X, y, partitions = data["X"], data["y"], data["partitions"]

        # get the data to be used for evaluation
        trn, tst = partitions[record["partition"]]

        model = copy.deepcopy(record["proto"])
        model.fit(X[trn], y[trn])

        trn_y_hat = model.predict(X[trn])
        tst_y_hat = model.predict(X[tst])

        # set the required entries including the prediction from the model
        record["trn_y_hat"] = trn_y_hat[:, 0]
        record["y_hat"] = tst_y_hat[:, 0]
        record["model"] = model
        record["success"] = True

        return model

    except Exception as e:
        record["success"] = False

        if show_warning:
            print("WARNING: %s has failed: %s" % (model, str(e)))

        if raise_exception:
            raise e



import numpy as np

try:
    from smt.surrogate_models import KRG
except:
    raise Exception("Model not found. Please execute: 'pip install smt'")

from ezmodel.core.model import Model


class smtKriging(Model):

    def __init__(self, theta0=1e-4, **kwargs) -> None:
        super().__init__(**kwargs)
        self.theta0 = theta0

    def _fit(self, X, y, **kwargs):
        n_var = X.shape[1]

        sm = KRG(theta0=[self.theta0] * n_var)
        sm.set_training_values(X, y)
        sm.train()

        self.model = sm

    def _predict(self, X, out, **kwargs):

        if "y" in out:
            out["y"] = self.model.predict_values(X)

        if "sigma" in out:
            var = self.model.predict_variances(X)

            var[var <= 0] = 0
            out["var"] = var
            out["sigma"] = np.sqrt(var)

    @classmethod
    def hyperparameters(cls):
        return {

        }



import numpy as np

from ezmodel.core.transformation import Transformation


class Standardization(Transformation):

    def __init__(self, mean=None, std=None) -> None:
        super().__init__()
        self.mean = mean
        self.std = std

    def forward(self, X):
        if self.mean is None:
            self.mean = np.mean(X, axis=0)

        if self.std is None:
            self.std = np.std(X, axis=0)

        val = (X - self.mean) / self.std
        return val

    def backward(self, X):
        return (X * self.std) + self.mean




import numpy as np


def create(func, n_train, n_test, n_var=1, seed=1):
    np.random.seed(seed)
    X = np.random.random((n_train, n_var)) * 2 * np.pi - np.pi
    y = func(X).sum(axis=1)[:, None]

    _X = np.random.random((n_test, n_var)) * 2 * np.pi - np.pi
    _y = func(_X).sum(axis=1)[:, None]

    return X, y, _X, _y


def sine_function(n_train, n_test):
    return create(lambda X: np.sin(5*X), n_train, n_test)


def constant_function(n_train, n_test, noise=0.0, **kwargs):
    return create(lambda X: np.random.randn(len(X))[:, None] * noise, n_train, n_test)


def square_function(n_train, n_test, noise=0.0, **kwargs):
    return create(lambda X: X ** 2 + np.random.randn(len(X))[:, None] * noise, n_train, n_test)


def linear_function(n_train, n_test, noise=0.0, **kwargs):
    return create(lambda X: X + np.random.randn(len(X))[:, None] * noise, n_train, n_test, **kwargs)



import numpy as np

from ezmodel.core.model import Model
from ezmodel.util.dist import calc_dist


class RBF(Model):

    def __init__(self,
                 kernel="linear",
                 tail="constant",
                 sigma=1.0,
                 rho=1e-6,
                 normalized=False,
                 optimize=False,
                 **kwargs) -> None:

        super().__init__(eliminate_duplicates=True, eliminate_duplicates_eps=1e-8, **kwargs)
        self.tail = tail
        self.rho = rho
        self.sigma = sigma
        self.normalized = normalized
        self.optimize = optimize

        if kernel == "linear":
            kernel = kernel_linear
        elif kernel == "quadratic":
            kernel = kernel_quadratic
        elif kernel == "cubic":
            kernel = kernel_cubic
        elif kernel == "gaussian":
            kernel = kernel_gaussian
        elif kernel == "mq":
            kernel = kernel_multi_quadr
        elif kernel == "tps":
            kernel = kernel_tps
        elif kernel == "periodic":
            kernel = kernel_periodic
        else:
            raise Exception("Unknown Kernel Function: " + kernel)

        self.kernel = kernel

    def _fit(self, X, y, **kwargs):

        rho, tail, kernel, sigma, normalized = self.rho, self.tail, self.kernel, self.sigma, self.normalized

        if self.optimize:

            sigmas = np.linspace(0.0001, 20, 30)
            models = [rbf_fit(X, y, kernel, sigma=sigma, tail=tail, rho=rho, normalized=normalized) for sigma in sigmas]
            f = np.array([model["loocv"] for model in models])
            cond = np.array([model["cond"] for model in models])
            f[cond > 1e12] = np.inf

            # _f = np.array([calc_loocv(X, y, kernel, sigma=sigma, tail=tail, rho=rho)[0] for sigma in sigmas])
            # assert np.all(np.logical_or(cond > 1e12, np.abs(f - _f) < 1e-6))

            k = f.argmin()

            print(sigmas[k])

            self.model = models[k]

            # print(np.all(np.logical_or(cond > 1e12, np.abs(f - _f) < 1e-6)))

        else:
            self.model = rbf_fit(X, y, kernel, tail=tail, rho=rho, sigma=sigma, normalized=normalized)

    def _predict(self, X, out):
        y_hat = rbf_predict(self.model, X)
        out["y"] = y_hat

    @classmethod
    def hyperparameters(cls):
        return {
            "kernel": ["linear", "cubic", "gaussian", "mq"],
            "tail": ["constant", "linear", "quadratic", "linear+quadratic"],
            # "rho": [None, 1e-3, 1e-2, 1e-1, 0.5, 1.0, 2.0],
            "normalized": [True, False],
        }


def rbf_kernel(X, phi, tail="linear", **kwargs):
    n, m = X.shape

    if tail is None:
        P = np.zeros((n, 0))
    if tail == "constant":
        P = np.ones((n, 1))
    elif tail == "linear":
        P = np.column_stack((np.ones(n), X))
    elif tail == "quadratic":
        P = np.column_stack((np.ones(n), X ** 2))
    elif tail == "linear+quadratic":
        P = np.column_stack((np.ones(n), X, X ** 2))

    return np.column_stack([phi, P])


def rbf_fit(X, y, func, Xp=None, rho=0.0, normalized=False, **kwargs):
    if Xp is None:
        Xp = X

    phi = calc_dist(X, Xp)
    phi = func(phi, **kwargs)

    phi_norm = 1.0
    if normalized:
        phi_norm = phi.sum(axis=0)[None, :]
        phi = phi / phi_norm

    if rho is not None:
        phi = phi + np.eye(len(phi)) * (rho ** 2)

    K = rbf_kernel(X, phi, **kwargs)
    n, m = K.shape

    lhs = np.zeros((m, m))
    lhs[:n, :m] = K
    lhs[n:, :n] = K[:, n:].T

    rhs = np.zeros((m, 1))
    rhs[:n] = y

    A_inv, cond = svd_inv(lhs)
    coef = A_inv @ rhs

    c, K = coef[:n, 0], A_inv[:n, :n]
    e = c / (np.diag(K) + 1e-128)
    loocv = (e ** 2).sum()
    gcv = (c ** 2).sum() / (np.diag(K).mean() ** 2)

    return dict(X=X, cond=cond, e=e, loocv=loocv, gcv=gcv, coef=coef, func=func, phi_norm=phi_norm, kwargs=kwargs)


def rbf_predict(model, X):
    phi = model["func"](calc_dist(X, model["X"]), **model["kwargs"])
    phi = phi / model["phi_norm"]
    phi = rbf_kernel(X, phi, **model["kwargs"])

    y = phi @ model["coef"]
    return y


def kernel_linear(r, sigma=1.0, **kwargs):
    return sigma * r


def kernel_quadratic(r, sigma=1.0, **kwargs):
    return (sigma * r) ** 2


def kernel_cubic(r, sigma=1.0, **kwargs):
    return (sigma * r) ** 3


def kernel_gaussian(r, sigma=None, **kwargs):
    return np.exp(- (sigma * r ** 2))


def kernel_periodic(r, sigma=1.0, **kwargs):
    return (sigma ** 2) * np.exp(- 2 * np.sin((np.pi * r) / 5) ** 2)


def kernel_multi_quadr(r, sigma=1.0, **kwargs):
    return ((r ** 2) + (sigma ** 2)) ** 0.5


def kernel_tps(r, **kwargs):
    r[r < np.finfo(float).eps] = np.finfo(float).eps
    return (r ** 2) * np.log(r)


def svd_inv(A):
    U, S, V = np.linalg.svd(A)
    Ainv = V.T @ np.diag(1 / S) @ U.T
    cond = np.abs(np.max(S)) / np.abs(np.min(S))
    return Ainv, cond


def calc_loocv(X, y, *args, **kwargs):
    e = []

    for k in range(len(X)):
        trn = [i for i in range(len(X)) if i != k]
        tst = [k]

        model = rbf_fit(X[trn], y[trn], *args, **kwargs)
        y_hat = rbf_predict(model, X[tst])[:, 0]
        y_true = y[tst, 0]

        e.append(y_true - y_hat)

    e = np.array(e)

    return (e ** 2).sum(), e



import numpy as np

from ezmodel.core.transformation import Transformation


class Plog(Transformation):

    def __init__(self) -> None:
        super().__init__()

    def forward(self, y):
        yp = np.zeros_like(y)

        larger = y >= 0
        yp[larger] = np.log(1 + y[larger])
        yp[~larger] = - np.log(1 - y[~larger])

        return yp

    def backward(self, yp):
        y = np.zeros_like(yp)

        larger = yp >= 0
        y[larger] = np.exp(yp[larger]) - 1
        y[~larger] = 1 - np.exp(-yp[~larger])

        return y



import copy
from abc import abstractmethod

from ezmodel.util.misc import dict_to_str


class ModelFactory:

    @abstractmethod
    def do(self):
        pass


class ModelFactoryByClazz(ModelFactory):

    def __init__(self, clazz, params=None, create_instance=True, **kwargs) -> None:
        super().__init__()
        self.clazz = clazz
        self.kwargs = kwargs
        self.params = params
        self.create_instance = create_instance

    def do(self):
        params, clazz, kwargs = self.params, self.clazz, self.kwargs
        if params is None:
            params = self.clazz.hyperparameters()

        ret = {}
        for vals in dfs(params):
            label = f"{clazz.__name__}[{dict_to_str(vals)}]"

            obj = None
            if self.create_instance:
                all_kwargs = {**kwargs, **vals}
                obj = clazz(**all_kwargs)

            ret[label] = dict(label=label, clazz=clazz, params=vals, defaults=kwargs, model=obj)

        return ret


def models_from_clazzes(*clazzes, **defaults):
    models = {}
    for clazz in clazzes:
        models = {**models, **ModelFactoryByClazz(clazz, **defaults).do()}
    return models


# --------------------------------------------------------------
# Util
# --------------------------------------------------------------


def dfs(params):
    """

    Parameters
    ----------
    params : dict
    A dictionary where each key has a list of values.

    Returns
    -------
    comb : list
    All possible combinations of this dictionary having selected
    one entry from each key.

    """
    ret = []
    dfs_rec({}, params, ret)
    return ret


def dfs_rec(entry, params, ret):
    if len(params) == 0:
        ret.append(entry)
    else:
        _params = copy.deepcopy(params)
        key = list(params.keys())[0]
        del _params[key]

        for value in params[key]:
            _comb = copy.deepcopy(entry)
            _comb[key] = value

            dfs_rec(_comb, _params, ret)



import numpy as np

from pymoo.algorithms.moo.nsga2 import RankAndCrowdingSurvival
from pymoo.core.population import Population
from pymoo.core.problem import Problem
from pymoo.util.clearing import EpsilonClearing
from pymoo.util.misc import vectorized_cdist, norm_eucl_dist_by_bounds
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting


class Selection:

    def __init__(self, **kwargs) -> None:
        super().__init__()
        for k, v in kwargs.items():
            self.__dict__[k] = v


class MinSelection(Selection):

    def do(self, rem):
        return rem[self.F[rem].argmin()]


class RandomSelection(Selection):

    def do(self, rem):
        return rem[np.random.randint(len(rem))]


class MinMaxSelection(Selection):

    def __init__(self, min_eps=0.01, **kwargs) -> None:
        super().__init__(min_eps=min_eps, **kwargs)

    def do(self, rem):
        F, D = self.F[rem], self.D[rem][:, rem]

        _min = F.argmin()
        clearing = EpsilonClearing(D, self.min_eps)
        clearing.select(_min)

        _rem = clearing.remaining()
        if len(_rem) > 0:
            _max = _rem[F[_rem].argmax()]
            return [rem[_min], rem[_max]]
        else:
            return rem[_min]


class FrontwiseSelection(Selection):

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        F = kwargs.get("F")
        G = kwargs.get("G")

        class MyProblem(Problem):
            def __init__(self, **kwargs):
                super().__init__(n_constr=0 if G is None else 1, **kwargs)

        pop = Population.new(index=np.arange(len(F)), F=F)

        if G is not None:
            pop.set("G", G)

        pop = RankAndCrowdingSurvival(nds=NonDominatedSorting()).do(MyProblem(), pop, n_survive=len(pop))

        self.rank = pop.get("rank")[pop.get("index")]
        self.crowding = pop.get("crowding")[pop.get("index")]

    def do(self, rem):
        _rank = self.rank[rem]
        _crowding = self.crowding[rem]
        I = np.lexsort([- _crowding, _rank])
        return rem[I[0]]


def aggregate_by_eps_clearing(X,
                              eps,
                              selection=MinSelection,
                              return_cluster=False,
                              func_dist=vectorized_cdist,
                              func_dist_by_bounds=norm_eucl_dist_by_bounds,
                              calc_distance=True,
                              problem=None,
                              xl=None,
                              xu=None,
                              **kwargs):
    if calc_distance:
        if problem is None:
            D = func_dist(X, X)
        else:
            if problem is not None:
                xl, xu = problem.bounds()
            D = func_dist_by_bounds(X, X, xl, xu)
    else:
        D = X

    clearing = EpsilonClearing(D, eps)

    sel = selection(problem=problem, X=X, D=D, **kwargs)

    D = {}

    while clearing.has_remaining():
        rem = clearing.remaining()
        S = sel.do(rem)

        if isinstance(S, list):
            for s in S:
                cleared = clearing.select(s)
                D[s] = cleared
        else:
            cleared = clearing.select(S)
            D[S] = cleared

    I = np.array(list(D.keys()))

    if return_cluster:
        return I, D
    else:
        return I



from ezmodel.core.benchmark import Benchmark
from ezmodel.core.factory import models_from_clazzes
from ezmodel.core.partitioning import merge_and_partition
from ezmodel.core.selection import ModelSelection
from ezmodel.models.kriging import Kriging
from ezmodel.models.rbf import RBF


def fit(
        X,
        y,
        X_test=None,
        y_test=None,
        partitions=None,
        clazzes=[Kriging, RBF],
        defaults={},
        models=None,
        return_benchmark=False,
        return_selection=False,
        **kwargs):
    """

    Find the best out of a variety of surrogates regarding a metric.

    Parameters
    ----------
    X : np.array
        The design space data.

    y : np.array
        The target value to model

    X_test : optional, np.array
        If no partitioning but a test set should be used this shall be set

    y_test : optional, np.array
        If no partitioning but a test set should be used this shall be set

    partitions : list
        The partitions to be used to find the best surrogate

    clazzes : list
        class instances which are used to create the models

    defaults : dict
        If the classes should be instantiated by default values they should be set here

    models : list
        The models can also be provided directly. Then clazzes and defaults can be `None`

    return_benchmark : bool
        Whether the benchmark shall finally be return too or not

    Returns
    -------

    model : Model
        The best model found on the data.

    benchmark : Benchmark
        The benchmark used to find the best model.


    """

    if X_test is not None and y_test is not None:
        X, y, partitions = merge_and_partition((X, y), (X_test, y_test))
        refit = False
    else:

        # by default just do cross validation
        if partitions is None:
            raise Exception("Either provide partitions or a test set.")

        refit = True

    if models is None:
        models = models_from_clazzes(*clazzes, **defaults)

    benchmark = Benchmark(models)

    selection = ModelSelection(benchmark, refit=refit, **kwargs)
    model = selection.do(X, y, partitions)

    ret = [model]

    if return_benchmark:
        ret.append(benchmark)
    if return_selection:
        ret.append(selection)

    if len(ret) == 1:
        return ret[0]
    else:
        return tuple(ret)



import numpy as np

from ezmodel.core.transformation import Transformation


class ZeroToOneNormalization(Transformation):

    def __init__(self, xl=None, xu=None, estimate_bounds=True) -> None:
        super().__init__()
        self.xl = xl
        self.xu = xu
        self.estimate_bounds = estimate_bounds

    def forward(self, X):

        if self.estimate_bounds:
            if self.xl is None:
                self.xl = np.min(X, axis=0)
            if self.xu is None:
                self.xu = np.max(X, axis=0)

        xl, xu = self.xl, self.xu

        # if np.any(xl == xu):
        #     raise Exception("Normalization failed because lower and upper bounds are equal!")

        # calculate the denominator
        denom = xu - xl

        # we can not divide by zero -> plus small epsilon
        denom += (denom == 0) * 1e-32

        # normalize the actual values
        N = (X - xl) / denom

        return N

    def backward(self, X):
        return X * (self.xu - self.xl) + self.xl



import random

import numpy as np


class Partitioning:

    def __init__(self, seed=None, include_trn_in_test=False) -> None:
        super().__init__()
        self.seed = seed
        self.include_trn_in_test = include_trn_in_test

    def do(self, X):

        # set the random seed to ensure reproducibility
        if self.seed is not None:
            random.seed(self.seed)
            np.random.seed(self.seed)

        ret = self._do(X)

        if self.include_trn_in_test:
            for k, (trn, tst) in enumerate(ret):
                ret[k] = (trn, trn + tst)

        return ret

    def _do(self, X):
        pass


def merge_and_partition(trn, tst):
    X, y = trn
    X_test, y_test = tst
    partitions = [(np.arange(len(X)), np.arange(len(X), len(X) + len(X_test)))]
    X, y = np.row_stack([X, X_test]), np.concatenate([y, y_test])
    return X, y, partitions



import matplotlib.pyplot as plt
import numpy as np

from ezmodel.models.idw import InverseDistanceWeighting
from ezmodel.util.sample_from_func import sine_function

idw = InverseDistanceWeighting()

# create some data to test this model on
X, y, _X, _y = sine_function(20, 1000)

# let the model fit the data
idw.fit(X, y)

# predict the data using the model
y_hat = idw.predict(_X)

# predict the data using the model
_X = _X[np.argsort(_X[:, 0])]
y_hat = idw.predict(_X)

plt.scatter(X, y, label="Data")
plt.plot(_X, y_hat, color="black", label="IDW")
plt.legend()
plt.show()



import math
import random

from ezmodel.core.partitioning import Partitioning


class CrossvalidationPartitioning(Partitioning):

    def __init__(self,
                 k_folds=5,
                 randomize=True,
                 **kwargs):

        super().__init__(**kwargs)
        self.randomize = randomize
        self.k_folds = k_folds

    def _do(self, X):
        n = len(X) if not isinstance(X, int) else X
        assert n > 1

        k_folds = min(self.k_folds, n)

        indices = list(range(n))
        if self.randomize:
            random.shuffle(indices)

        tst = [set() for _ in range(k_folds)]
        for k in range(n):
            tst[k % k_folds].add(indices[k])

        trn = [[j for j in indices if j not in tst[i]] for i in range(k_folds)]

        ret = []
        for _trn, _tst in zip(trn, tst):
            ret.append((_trn, list(_tst)))

        return ret



from abc import abstractmethod


class Transformation:

    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def forward(self, X):
        pass

    @abstractmethod
    def backward(self, X):
        pass


class NoNormalization(Transformation):

    def __init__(self) -> None:
        super().__init__()

    def forward(self, X):
        return X

    def backward(self, X):
        return X



from ezmodel.core.transformation import Transformation
import numpy as np


class StochasticTunneling(Transformation):

    def __init__(self, gamma, y_min=None) -> None:
        super().__init__()
        self.y_min = y_min
        self.gamma = gamma

    def forward(self, y):
        if self.y_min is None:
            self.y_min = y.min(axis=0)

        return 1 - np.exp(- self.gamma * (y - self.y_min))

    def backward(self, yp):
        yp = np.clip(yp, None, 1.0 - 1e-16)
        return self.y_min - np.log(1 - yp) / self.gamma


if __name__ == "__main__":
    F = np.random.random((100, 1))

    tunneling = StochasticTunneling(0.1)
    Fp = tunneling.forward(F)



import numpy as np

import pandas as pd
pd.set_option('display.expand_frame_repr', False)
pd.set_option('max_colwidth', 1000)

from ezmodel.core.benchmark import Benchmark
from ezmodel.core.factory import models_from_clazzes
from ezmodel.models.kriging import Kriging
from ezmodel.models.rbf import RBF
from ezmodel.util.partitioning.crossvalidation import CrossvalidationPartitioning

X = np.random.random((100, 3)) * 2 * np.pi
y = np.sin(X).sum(axis=1)

models = models_from_clazzes(RBF, Kriging)

# set up the benchmark and add the models to be used
benchmark = Benchmark(models, n_threads=4, verbose=True, raise_exception=True)

# create partitions to validate the performance of each model
partitions = CrossvalidationPartitioning(k_folds=5, seed=1).do(X)

# runs the experiment with the specified partitioning
benchmark.do(X, y, partitions=partitions)

# print out the benchmark results
print(benchmark.statistics("mae", sort="ascending"))



import numpy as np

try:
    from sklearn.linear_model import LinearRegression
    from sklearn.pipeline import make_pipeline
    from sklearn.preprocessing import PolynomialFeatures, StandardScaler
except:
    raise Exception("Model not found. Please execute: 'pip install sklearn'")

from ezmodel.core.model import Model


class PolynomialRegression(Model):

    def __init__(self, degree=3, fail_if_not_enough_points=True, **kwargs) -> None:
        super().__init__(**kwargs)
        self.degree = degree
        self.model = None
        self.fail_if_not_enough_points = fail_if_not_enough_points

    def _fit(self, X, y, **kwargs):

        n_min_points = PolynomialFeatures(self.degree).fit_transform(X).shape[1]

        if not self.fail_if_not_enough_points or len(X) >= n_min_points:
            model = make_pipeline(PolynomialFeatures(self.degree), StandardScaler(), LinearRegression())
            model.fit(X, y[:, 0])
            self.model = model
        else:
            raise Exception(f"For Polynomial Regression of degree {self.degree} at least {n_min_points} are necessary.")

    def _predict(self, X, out):
        out["y"] = self.model.predict(X)[:, None]

    @classmethod
    def hyperparameters(cls):
        return {
            "degree": np.arange(1, 4)
        }



import sys
from inspect import getmembers

import numpy as np
from pydacefit.dace import DACE

from ezmodel.core.model import Model


def get_corr(corr):
    for name, func in getmembers(sys.modules["pydacefit.corr"]):
        if name == "corr_" + corr:
            return func
    raise Exception("Correlation not found.")


def get_regr(corr):
    for name, func in getmembers(sys.modules["pydacefit.regr"]):
        if name == "regr_" + corr:
            return func
    raise Exception("Regression not found not found.")


class Kriging(Model):

    def __init__(self,
                 regr="linear",
                 corr="gauss",
                 ARD=False,
                 theta=1.0,
                 thetaL=0.00001,
                 thetaU=100.0,
                 **kwargs) -> None:

        super().__init__(eliminate_duplicates=True, **kwargs)
        self.regr = regr
        self.corr = corr
        self.ARD = ARD
        self.theta = theta
        self.thetaL = thetaL
        self.thetaU = thetaU

    def _fit(self, X, y, **kwargs):
        func_regr, func_corr = get_regr(self.regr), get_corr(self.corr)
        theta, thetaL, thetaU = self.theta, self.thetaL, self.thetaU

        if self.ARD:
            if self.thetaL is None or self.thetaU is None:
                pass
                # raise Exception("Bounds of theta must be given for Automatic Relevance Detection (ARD)!")
            else:
                _, m = X.shape
                theta = np.full(m, theta)
                thetaL = np.full(m, thetaL)
                thetaU = np.full(m, thetaU)

        self.model = DACE(regr=func_regr, corr=func_corr, theta=theta, thetaL=thetaL, thetaU=thetaU)
        self.model.fit(X, y)

    def _predict(self, X, out, **kwargs):
        calc_sigma = "sigma" in out
        calc_variance = "var" in out

        return_mse = calc_variance or calc_sigma

        ret = self.model.predict(X, return_mse=return_mse, **kwargs)

        if not return_mse:
            out["y"] = ret
        else:
            out["y"], var = ret
            var[var <= 0] = 0
            out["var"], out["sigma"] = var, np.sqrt(var)

    @classmethod
    def hyperparameters(cls):
        return {
            "regr": ["constant", "linear"],
            "corr": ["gauss", "cubic", "exp"],
            "thetaU": [20, 100],
            "ARD": [False, True]
        }



import inspect
import sys

import numpy as np
from scipy.stats import spearmanr
from sklearn.metrics import r2_score

METRICS = ["mse", "mae", "r2", "spear"]


def calc_metric(metric, y, y_hat):
    check_equal_shape(y, y_hat)

    funcs = dict(inspect.getmembers(sys.modules[__name__]))

    if metric not in funcs:
        raise Exception("Metric is not known.")
    else:
        return funcs[metric](y, y_hat)


# --------------------------------------------------------------
# Metrics
# --------------------------------------------------------------

def mse(y, y_hat):
    return np.mean((y - y_hat) ** 2)


def mae(y, y_hat):
    return np.mean(np.abs(y - y_hat))


def r2(y, y_hat):
    return r2_score(y, y_hat)


def spear(y, y_hat):
    return spearmanr(y, y_hat).correlation


# --------------------------------------------------------------
# Util
# --------------------------------------------------------------

def check_equal_shape(a, b):
    assert a.shape == b.shape





import matplotlib.pyplot as plt
import numpy as np
from ezmodel.models.smt import smtKriging
from ezmodel.util.metrics import calc_metric
from ezmodel.util.sample_from_func import sine_function

model = smtKriging()

# create some data to test this model on
X, y, _X, _y = sine_function(100, 20)

# let the model fit the data
model.fit(X, y)

# predict the data using the model
_X = _X[np.argsort(_X[:, 0])]
y_hat, y_var = model.predict(_X, return_values_of=["y", "var"])

# calculate the error and print it
mae = calc_metric("mae", _y, y_hat)
print("Mean Absolute Error", mae)

plt.scatter(X, y, label="Data")
plt.plot(_X, y_hat, color="black", label="RBF")
plt.legend()
plt.show()




from ezmodel.core.model import Model
from ezmodel.experimental.custom.kernel import dist


class InverseDistanceWeighting(Model):

    def __init__(self, p=2.0, eps=1e-32, **kwargs) -> None:
        super().__init__(**kwargs)
        self.p = p
        self.eps = eps

    def _fit(self, X, y, **kwargs):
        pass

    def _predict(self, X, out, **kwargs):
        _X, _y = self.X, self.y[:, 0]

        D = dist(X, _X)
        D[D <= self.eps] = self.eps

        w = 1 / D ** self.p

        for k, d in enumerate(D):
            is_zero = (d <= self.eps)
            if is_zero.sum() > 0:
                w[k, ~is_zero] = 0
                w[k, is_zero] = 1

        w = w / w.sum(axis=1)[:, None]

        y = (_y * w).sum(axis=1)

        out["y"] = y[:, None]

    @classmethod
    def hyperparameters(cls):
        return {}



import math

import numpy as np

from ezmodel.core.partitioning import Partitioning


class RandomPartitioning(Partitioning):

    def __init__(self,
                 perc_train=0.7,
                 n_sets=1,
                 **kwargs):
        super().__init__(**kwargs)
        self.n_sets = n_sets
        self.perc_train = perc_train

    def _do(self, X):
        n = len(X) if not isinstance(X, int) else X

        n_train = math.ceil(self.perc_train * n)

        partitions = [rnd(n, n_train) for _ in range(self.n_sets)]

        return partitions


def rnd(n, k):
    M = np.random.permutation(n)
    return M[:k], M[k:]



from copy import copy

from ezmodel.core.benchmark import Benchmark
from ezmodel.core.model import Model


class ModelSelection(Model):

    def __init__(self,
                 benchmark_or_models,
                 refit=True,
                 max_trn_error=None,
                 sorted_by=None,  # for instance ("mse", "mean", True)
                 ) -> None:

        super().__init__()

        if not isinstance(benchmark_or_models, Benchmark):
            self.benchmark = Benchmark(benchmark_or_models)
        else:
            self.benchmark = benchmark_or_models

        self.max_trn_error = max_trn_error
        self.sorted_by = sorted_by

        self.refit = refit

        self.data = None
        self.best = None
        self.model = None

    def do(self, X, y, *args, return_result=False, **kwargs):

        if self.benchmark.data is None:
            self.benchmark.do(X, y, *args, **kwargs)

        R = self.benchmark.results(only_successful=True,
                                   as_list=True,
                                   sorted_by=self.sorted_by,
                                   max_trn_error=self.max_trn_error)

        if len(R) == 0:
            raise Exception("No model could have been fitted successfully!")

        best = R[0]

        if self.refit:
            model = copy(best["model"])
            model.fit(X, y)
        else:
            if best["n_runs"] > 1:
                raise Exception("refit is necessary of the benchmark has multiple partitions!")
            else:
                model = best["runs"][0]["model"]

        self.data = R
        self.best = best
        self.model = model

        if not return_result:
            return model
        else:
            return model, best

    def statistics(self):
        if self.data is None:
            return None
        else:
            import pandas as pd
            data = [(e["label"], e["metric"]) for e in self.data]
            df = pd.DataFrame(data, columns=["label", self.metric])
            return df

    def fit(self, X, y, **kwargs):
        if self.model is None:
            self.do(X, y, **kwargs)
        else:
            self.model.fit(X, y, **kwargs)
        return self.model

    def predict(self, X, **kwargs):
        return self.model.predict(X, **kwargs)



import time

import numpy as np

from ezmodel.core.transformation import NoNormalization
from ezmodel.util.misc import is_duplicate, at_least2d


class Model:

    def __init__(self,
                 norm_X=NoNormalization(),
                 norm_y=NoNormalization(),
                 active_dims=None,
                 filter_nan_and_inf=True,
                 eliminate_duplicates=False,
                 eliminate_duplicates_eps=1e-16,
                 raise_exception_while_fitting=True,
                 raise_exception_while_prediction=True,
                 verbose=False,
                 **kwargs):

        self.norm_X = norm_X
        self.norm_y = norm_y

        self.eliminate_duplicates = eliminate_duplicates
        self.eliminate_duplicates_eps = eliminate_duplicates_eps

        self.active_dims = active_dims
        self.filter_nan_and_inf = filter_nan_and_inf
        self.verbose = verbose

        self.time = None
        self.model = None
        self.X, self.y = None, None
        self._X, self._y = None, None
        self.success = None
        self.data = {}

        self.raise_exception_while_fitting = raise_exception_while_fitting
        self.raise_exception_while_prediction = raise_exception_while_prediction
        self.exception = None

        self.has_been_fitted = False

    def preprocess(self, X, y, **kwargs):

        if self.active_dims is not None:
            X = X[:, self.active_dims]

        if self.eliminate_duplicates:
            I = ~is_duplicate(X, eps=self.eliminate_duplicates_eps)
            X, y = X[I], y[I]

        if self.filter_nan_and_inf:
            X_I = np.all(~np.isnan(X) & ~np.isinf(X), axis=1)
            y_I = np.all(~np.isnan(y) & ~np.isinf(y), axis=1)
            X, y = X[X_I & y_I], y[X_I & y_I]

        X, y = self.norm_X.forward(X), self.norm_y.forward(y)
        X, y = self._preprocess(X, y, **kwargs)

        return X, y

    def postprocess(self, out, **kwargs):
        if "y" in out:
            out["y"] = self.norm_y.backward(out["y"])
        out = self._postprocess(out, **kwargs)
        return out

    def fit(self, X, y, **kwargs):
        X, y = at_least2d(X, expand="r"), at_least2d(y, expand="c")
        assert len(X) == len(y)
        self._X, self._y = X, y

        self.X, self.y = self.preprocess(X, y)

        start = time.time()

        try:

            # fit the model given the data
            self._fit(self.X, self.y, **kwargs)

            # do some parameter optimization if the model requires it
            self._optimize(**kwargs)

            # if no exception occurs set the model to be fitted successfully
            self.success = True

        except Exception as ex:
            self.success = False
            self.exception = ex
            if self.raise_exception_while_fitting:
                raise ex

        self.has_been_fitted = True
        self.time = time.time() - start

        return self

    def predict(self, X,
                return_values_of=["y"],
                return_as_dictionary=False,
                **kwargs):

        if not self.success:
            if self.raise_exception_while_fitting:
                raise Exception("There was an error while fitting the model.")
            else:
                return np.full(len(X), np.nan)

        if self.active_dims is not None:
            X = X[:, self.active_dims]

        # normalize the input
        X = self.norm_X.forward(at_least2d(X, expand="r"))

        # write in the output dictionary what should be returned
        out = {}
        for k in return_values_of:
            out[k] = None

        try:
            # get the prediction from the actual implementation
            self._predict(X, out, **kwargs)

            # do the post processing of the outputs
            out = self.postprocess(out, **kwargs)

        except Exception as e:
            if self.raise_exception_while_prediction:
                raise e
            else:
                out["y"] = np.full(len(X), np.inf)

        if return_as_dictionary:
            return out
        else:
            ret = tuple([out[v] for v in return_values_of])
            return ret if len(ret) > 1 else ret[0]

    @classmethod
    def hyperparameters(cls):
        return {}

    def _preprocess(self, X, y, **kwargs):
        return X, y

    def _postprocess(self, out, **kwargs):
        return out

    def _fit(self, X, y, **kwargs):
        pass

    def _predict(self, X, out, **kwargs):
        pass

    def _optimize(self, **kwargs):
        pass



import setuptools
from setuptools import setup

from ezmodel.version import __version__

__name__ = "ezmodel"
__author__ = "Julian Blank"
__url__ = ""

data = dict(
    name=__name__,
    version=__version__,
    author=__author__,
    url=__url__,
    python_requires='>=3.6',
    author_email="blankjul@msu.edu",
    description="Machine Learning, Model, Surrogate, Metamodels, Response Surface",
    license='Apache License 2.0',
    keywords="model",
    install_requires=['numpy', "pandas", "scipy", "sklearn", "pydacefit", "matplotlib"],
    include_package_data=True,
    platforms='any',
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Mathematics'
    ]
)


# ---------------------------------------------------------------------------------------------------------
# OTHER METADATA
# ---------------------------------------------------------------------------------------------------------


# update the readme.rst to be part of setup
def readme():
    with open('README.rst') as f:
        return f.read()


def packages():
    return ["ezmodel"] + ["ezmodel." + e for e in setuptools.find_packages(where='ezmodel')]


data['long_description'] = readme()
data['long_description_content_type'] = 'text/x-rst'
data['packages'] = packages()

# ---------------------------------------------------------------------------------------------------------
# SETUP
# ---------------------------------------------------------------------------------------------------------

setup(**data)



try:
    import GPy
except:
    raise Exception("Model not found. Please execute: 'pip install GPy'")

import numpy as np

from ezmodel.core.model import Model

kernels = {
    "matern32": GPy.kern.Matern32,
    "matern52": GPy.kern.Matern52,
    "cosine": GPy.kern.Cosine,
    "ratquad": GPy.kern.RatQuad,
    "exponential": GPy.kern.Exponential,
}


def get_kernel(name, n_var, ARD):
    if name not in kernels:
        raise Exception("Kernel not found.")
    else:
        return kernels[name](n_var, ARD=ARD)


class gpyGP(Model):

    def __init__(self, kernel="matern52", ARD=False, optimizer='lbfgs', **kwargs) -> None:
        super().__init__(**kwargs)
        self.kernel = kernel
        self.ARD = ARD
        self.optimizer = optimizer

    def _fit(self, X, y, **kwargs):
        kernel = get_kernel(self.kernel, X.shape[1], self.ARD)
        model = GPy.models.GPRegression(X, y, kernel, normalizer=True)

        model.constrain_positive('')
        (kern_variance, kern_lengthscale, gaussian_noise) = model.parameter_names()

        model[kern_variance].constrain_bounded(1e-6, 1e6, warning=False)
        model[kern_lengthscale].constrain_bounded(1e-6, 1e6, warning=False)
        model[gaussian_noise].constrain_fixed(1e-6, warning=False)

        if self.optimizer == 'lbfgs':
            model.optimize_restarts(optimizer='lbfgs',
                                    num_restarts=10,
                                    num_processes=1,
                                    verbose=False)
        elif self.optimizer == 'ga':
            model.optimize(optimizer='ga')
        else:
            raise Exception("Unknown Optimizer!")

        self.model = model

    def _predict(self, X, out):
        if "sigma" in out:
            out["y"], var = self.model.predict(X)

            var[var <= 0] = 0
            out["var"] = var
            out["sigma"] = np.sqrt(var)
        else:
            out["y"], _ = self.model.predict(X)

    @classmethod
    def hyperparameters(cls):
        return {
            "kernel": ["matern52", "ratquad", "exponential"],
            "ARD": [False, True],
        }



from ezmodel.core.model import Model

try:
    import gpflow
except:
    raise Exception("Model not found. Please execute: 'pip install gpflow'")


class GPFlow(Model):

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def _fit(self, X, y, **kwargs):
        kernel = gpflow.kernels.SquaredExponential()
        m = gpflow.models.GPR(data=(X, y),
                              kernel=kernel,
                              mean_function=gpflow.mean_functions.Constant(),
                              noise_variance=1e-03)

        opt = gpflow.optimizers.Scipy()
        opt.minimize(m.training_loss, m.trainable_variables, options=dict(maxiter=100))
        self.model = m

    def _predict(self, X, out, **kwargs):
        mean, var = self.model.predict_f(X)

        out["y"] = mean.numpy()
        if "var" in out:
            out["var"] = var.numpy()



import numpy as np

from ezmodel.core.model import Model


class SimpleMean(Model):

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def _fit(self, _, y, **kwargs):
        self.model = np.mean(y, axis=0)

    def _predict(self, X, out):
        out["y"] = np.full((len(X), 1), self.model)

    @classmethod
    def hyperparameters(cls):
        return {}





import matplotlib.pyplot as plt
import numpy as np

from ezmodel.core.factory import models_from_clazzes
from ezmodel.core.selection import ModelSelection
from ezmodel.fit import fit
from ezmodel.models.kriging import Kriging
from ezmodel.models.rbf import RBF
from ezmodel.util.partitioning.random import RandomPartitioning
from ezmodel.util.sample_from_func import square_function

X, y, _X, _y = square_function(20, 200, n_var=1)

# select the best model from the given options
models = models_from_clazzes(RBF, Kriging)

model = fit(X, y, partitions=RandomPartitioning().do(X))

# predict the data using the model
y_hat = model.predict(_X)

# predict the data using the model
_X = _X[np.argsort(_X[:, 0])]
y_hat = model.predict(_X)

plt.scatter(X, y, label="Data")
plt.plot(_X, y_hat, color="black", label="Regression")
plt.legend()
plt.show()





try:
    from sklearn.pipeline import make_pipeline
    from sklearn.preprocessing import StandardScaler
    from sklearn.svm import SVR as _SVR
except:
    raise Exception("Model not found. Please execute: 'pip install sklearn'")


from ezmodel.core.model import Model


class SVR(Model):

    def __init__(self, kernel="rbf", eps=0.1, C=1.0, **kwargs) -> None:
        super().__init__(**kwargs)
        self.kernel = kernel
        self.eps = eps
        self.C = C

    def _fit(self, X, y, **kwargs):
        svr = _SVR(kernel=self.kernel, epsilon=self.eps, C=self.C, gamma='scale', degree=3, tol=0.001, shrinking=True)
        regr = make_pipeline(StandardScaler(), svr)
        regr.fit(X, y[:, 0])
        self.model = regr

    def _predict(self, X, out):
        out["y"] = self.model.predict(X)[:, None]

    @classmethod
    def hyperparameters(cls):
        return {
            "kernel": ['linear', 'poly', 'rbf', 'sigmoid'],
            "eps": [0.1, 0.01],
            "C": [0.5, 1.0, 2.0]
        }
